@@ -36,6 +36,12 @@
 #include "server.h"
 #include "sblist.h"
 
+/* timeout in microseconds on resource exhaustion to prevent excessive
+   cpu usage. */
+#ifndef FAILURE_TIMEOUT
+#define FAILURE_TIMEOUT 64
+#endif
+
 #ifndef MAX
 #define MAX(x, y) ((x) > (y) ? (x) : (y))
 #endif
@@ -105,6 +111,15 @@ struct thread {
 static void dolog(const char* fmt, ...) { }
 #endif
 
+static struct addrinfo* addr_choose(struct addrinfo* list, union sockaddr_union* bind_addr) {
+	int af = SOCKADDR_UNION_AF(bind_addr);
+	if(af == AF_UNSPEC) return list;
+	struct addrinfo* p;
+	for(p=list; p; p=p->ai_next)
+		if(p->ai_family == af) return p;
+	return list;
+}
+
 static int connect_socks_target(unsigned char *buf, size_t n, struct client *client) {
 	if(n < 5) return -EC_GENERAL_FAILURE;
 	if(buf[0] != 5) return -EC_GENERAL_FAILURE;
@@ -140,7 +155,8 @@ static int connect_socks_target(unsigned char *buf, size_t n, struct client *cli
 	port = (buf[minlen-2] << 8) | buf[minlen-1];
 	/* there's no suitable errorcode in rfc1928 for dns lookup failure */
 	if(resolve(namebuf, port, &remote)) return -EC_GENERAL_FAILURE;
-	int fd = socket(remote->ai_addr->sa_family, SOCK_STREAM, 0);
+	struct addrinfo* raddr = addr_choose(remote, &bind_addr);
+	int fd = socket(raddr->ai_family, SOCK_STREAM, 0);
 	if(fd == -1) {
 		eval_errno:
 		if(fd != -1) close(fd);
@@ -165,9 +181,10 @@ static int connect_socks_target(unsigned char *buf, size_t n, struct client *cli
 			return -EC_GENERAL_FAILURE;
 		}
 	}
-	if(SOCKADDR_UNION_AF(&bind_addr) != AF_UNSPEC && bindtoip(fd, &bind_addr) == -1)
+	if(SOCKADDR_UNION_AF(&bind_addr) == raddr->ai_family &&
+	   bindtoip(fd, &bind_addr) == -1)
 		goto eval_errno;
-	if(connect(fd, remote->ai_addr, remote->ai_addrlen) == -1)
+	if(connect(fd, raddr->ai_addr, raddr->ai_addrlen) == -1)
 		goto eval_errno;
 
 	freeaddrinfo(remote);
@@ -256,7 +273,6 @@ static void copyloop(int fd1, int fd2) {
 		   when a connection is really unused. */
 		switch(poll(fds, 2, 60*15*1000)) {
 			case 0:
-				send_error(fd1, EC_TTL_EXPIRED);
 				return;
 			case -1:
 				if(errno == EINTR || errno == EAGAIN) continue;
@@ -439,14 +455,19 @@ int main(int argc, char** argv) {
 		struct thread *curr = malloc(sizeof (struct thread));
 		if(!curr) goto oom;
 		curr->done = 0;
-		if(server_waitclient(&s, &c)) continue;
+		if(server_waitclient(&s, &c)) {
+			dolog("failed to accept connection\n");
+			free(curr);
+			usleep(FAILURE_TIMEOUT);
+			continue;
+		}
 		curr->client = c;
 		if(!sblist_add(threads, &curr)) {
 			close(curr->client.fd);
 			free(curr);
 			oom:
 			dolog("rejecting connection due to OOM\n");
-			usleep(16); /* prevent 100% CPU usage in OOM situation */
+			usleep(FAILURE_TIMEOUT); /* prevent 100% CPU usage in OOM situation */
 			continue;
 		}
 		pthread_attr_t *a = 0, attr;
